package per.goweii.anylayer;

import android.support.annotation.NonNull;
import android.support.annotation.Nullable;
import android.view.KeyEvent;
import android.view.View;
import android.view.ViewGroup;
import android.view.ViewTreeObserver;

import per.goweii.anylayer.utils.Utils;

/**
 * 管理view的动态添加和移除
 * 这里有几个生命周期的回调：
 * {@link #onAttach()}
 * {@link #onDetach()}
 */
public final class ViewManager {

    private ViewGroup mParent = null;
    private View mChild = null;

    private LayerKeyListener mLayerKeyListener = null;
    private LayerGlobalFocusChangeListener mLayerGlobalFocusChangeListener = null;

    private View currentKeyView = null;

    private OnLifeListener mOnLifeListener = null;
    private OnKeyListener mOnKeyListener = null;

    public ViewManager() {
    }

    public void setParent(@NonNull ViewGroup parent) {
        mParent = parent;
    }

    public void setChild(@NonNull View child) {
        mChild = child;
    }

    @Nullable
    public ViewGroup getParent() {
        return mParent;
    }

    @Nullable
    public View getChild() {
        return mChild;
    }

    @NonNull
    public ViewGroup requireParent() {
        return Utils.requireNonNull(mParent, "还未设置parent");
    }

    @NonNull
    public View requireChild() {
        return Utils.requireNonNull(mChild, "还未设置child");
    }

    private void checkChildParent() {
        ViewGroup parent = (ViewGroup) mChild.getParent();
        if (parent != null && parent != mParent) {
            parent.removeView(mChild);
        }
    }

    public void attach() {
        Utils.requireNonNull(mParent, "必须设置parent");
        checkChildParent();
        if (!isAttached()) {
            onAttach();
        }
    }

    public void detach() {
        if (isAttached()) {
            onDetach();
        }
    }

    public boolean isAttached() {
        return mChild != null && mChild.getParent() != null;
    }

    public void setOnLifeListener(@Nullable OnLifeListener onLifeListener) {
        mOnLifeListener = onLifeListener;
    }

    public void setOnKeyListener(@Nullable OnKeyListener onKeyListener) {
        mOnKeyListener = onKeyListener;
    }

    /**
     * 添加到父View
     */
    private void onAttach() {
        mParent.addView(mChild);
        if (mOnKeyListener != null) {
            mChild.setFocusable(true);
            mChild.setFocusableInTouchMode(true);
            mLayerKeyListener = new LayerKeyListener();
            findFocusViewAndBindKeyListener();
            mLayerGlobalFocusChangeListener = new LayerGlobalFocusChangeListener();
            mChild.getViewTreeObserver().addOnGlobalFocusChangeListener(mLayerGlobalFocusChangeListener);
        }
        if (mOnLifeListener != null) {
            mOnLifeListener.onAttach();
        }
    }

    /**
     * 从父View移除
     */
    private void onDetach() {
        if (currentKeyView != null) {
            currentKeyView.setOnKeyListener(null);
            mLayerKeyListener = null;
            mChild.getViewTreeObserver().removeOnGlobalFocusChangeListener(mLayerGlobalFocusChangeListener);
            mLayerGlobalFocusChangeListener = null;
        }
        mParent.removeView(mChild);
        if (mOnLifeListener != null) {
            mOnLifeListener.onDetach();
        }
    }

    private void findFocusViewAndBindKeyListener() {
        Utils.onViewLayout(mChild, new Runnable() {
            @Override
            public void run() {
                currentKeyView = mChild.findFocus();
                if (currentKeyView != null) {
                    currentKeyView.setOnKeyListener(mLayerKeyListener);
                    return;
                }
                mChild.requestFocus();
                currentKeyView = mChild.findFocus();
                if (currentKeyView != null) {
                    currentKeyView.setOnKeyListener(mLayerKeyListener);
                }
            }
        });
    }

    private final class LayerGlobalFocusChangeListener implements ViewTreeObserver.OnGlobalFocusChangeListener {
        @Override
        public void onGlobalFocusChanged(View oldFocus, View newFocus) {
            if (currentKeyView != null) {
                currentKeyView.setOnKeyListener(null);
            }
            currentKeyView = mChild.findFocus();
            if (currentKeyView != null) {
                currentKeyView.setOnKeyListener(mLayerKeyListener);
                return;
            }
            View rootFocus = mChild.getRootView().findFocus();
            if (rootFocus != null) {
                return;
            }
            mChild.requestFocus();
            currentKeyView = mChild.findFocus();
            if (currentKeyView != null) {
                currentKeyView.setOnKeyListener(mLayerKeyListener);
            }
        }
    }

    private final class LayerKeyListener implements View.OnKeyListener {
        @Override
        public boolean onKey(View v, int keyCode, KeyEvent event) {
            if (!isAttached()) {
                return false;
            }
            if (mOnKeyListener == null) {
                return false;
            }
            return mOnKeyListener.onKey(keyCode, event);
        }
    }

    public interface OnLifeListener {
        void onAttach();

        void onDetach();
    }

    public interface OnKeyListener {
        boolean onKey(int keyCode, KeyEvent event);
    }
}